//---------------------------------------------------------------------------
// DensitySeedSurvival.cpp
//---------------------------------------------------------------------------
#include "DensitySeedSurvival.h"
#include "SimManager.h"
#include "TreePopulation.h"
#include "GridBase.h"
#include "ParsingFunctions.h"
#include "ModelMath.h"
#include "Plot.h"
#include <math.h>

///////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// April 12, 2005 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
clDensitySeedSurvival::clDensitySeedSurvival( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
     {
       try
       {
         strcpy( m_cNameString, "Density Dependent Seed Survival" );

         mp_fDensDepSteepness = NULL;
         mp_fDensDepSlope = NULL;
         mp_iSeedGridCode = NULL;

         //Allowed file types
         m_iNumAllowedTypes = 2;
         mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
         mp_iAllowedFileTypes[0] = parfile;
         mp_iAllowedFileTypes[1] = detailed_output;

         //Versions
         m_fVersionNumber = 1;
         m_fMinimumVersionNumber = 1;
       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clDensitySeedSurvival::clDensitySeedSurvival" );
         throw( stcErr );
       }
}

///////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// April 12, 2005 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
clDensitySeedSurvival::~clDensitySeedSurvival()
{
  delete[] mp_fDensDepSteepness;
  delete[] mp_fDensDepSlope;
  delete[] mp_iSeedGridCode;
}

///////////////////////////////////////////////////////////////////////////////
// GetData
// Edit history:
// -------------
// April 12, 2005 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////
void clDensitySeedSurvival::GetData( DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
    clPlot *p_oPlot = mp_oSimManager->GetPlotObject();
    floatVal * p_fTemp = NULL; //for getting species-specific values
    char cLabel[MAX_GRID_LABEL_SIZE];
    float fXTemp, fYTemp;
    short int iNumSpecies = p_oPop->GetNumberOfSpecies(), i;

    //Declare our arrays
    mp_fDensDepSteepness = new float[iNumSpecies];
    mp_fDensDepSlope = new float[iNumSpecies];
    mp_iSeedGridCode = new short int[iNumSpecies];

    //Set up our temp array - pre-load with this behavior's species
    p_fTemp = new floatVal[m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      p_fTemp[i].code = mp_iWhatSpecies[i];

    //Density dependence slope
    FillSpeciesSpecificValue( p_oElement, "es_densDepSlope", "es_ddsVal", p_fTemp, m_iNumBehaviorSpecies, p_oPop, true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fDensDepSlope[p_fTemp[i].code] = p_fTemp[i].val;

    //Density dependence steepness
    FillSpeciesSpecificValue( p_oElement, "es_densDepSteepness", "es_ddstVal", p_fTemp, m_iNumBehaviorSpecies, p_oPop, true );
    //Now transfer the values to the permanent array
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      mp_fDensDepSteepness[p_fTemp[i].code] = p_fTemp[i].val;

    //Fetch the seed grid
    mp_oSeedGrid = mp_oSimManager->GetGridObject( "Dispersed Seeds" );
    if ( mp_oSeedGrid == NULL )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clDensitySeedSurvival::GetData" );
      strcpy( stcErr.cMoreInfo, "Disperse behaviors must be used with density-dependent seed survival." );
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      throw( stcErr );
    }

    //Now get the data codes
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    {
      sprintf( cLabel, "%s%d", "seeds_", mp_iWhatSpecies[i] );
      mp_iSeedGridCode[mp_iWhatSpecies[i]] = mp_oSeedGrid->GetFloatDataCode( cLabel );
      if ( -1 == mp_iSeedGridCode[mp_iWhatSpecies[i]] )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clDensitySeedSurvival::GetData" );
        strcpy( stcErr.cMoreInfo, "Unexpected grid data return code." );
        throw( stcErr );
      }
    }

    //Calculate the cell areas
    //Normal cell area - straightforward
    m_fNormalCellArea = mp_oSeedGrid->GetLengthXCells() * mp_oSeedGrid->GetLengthYCells();

    //X and Y edges - get the length of the last cells in the X and Y direction
    // (might be the same as all the other cells)
    fXTemp = p_oPlot->GetXPlotLength() - (mp_oSeedGrid->GetLengthXCells() * floor(p_oPlot->GetXPlotLength() / mp_oSeedGrid->GetLengthXCells()));
    fXTemp = (fXTemp == 0 ? mp_oSeedGrid->GetLengthXCells() : fXTemp);
    fYTemp = p_oPlot->GetYPlotLength() - (mp_oSeedGrid->GetLengthYCells() * floor(p_oPlot->GetYPlotLength() / mp_oSeedGrid->GetLengthYCells()));
    fYTemp = (fYTemp == 0 ? mp_oSeedGrid->GetLengthYCells() : fYTemp);

    m_fXEdgeCellArea = fXTemp * mp_oSeedGrid->GetLengthYCells();
    m_fYEdgeCellArea = fYTemp * mp_oSeedGrid->GetLengthXCells();
    m_fXYEdgeCellArea = fXTemp * fYTemp;

    delete[] p_fTemp;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDensitySeedSurvival::GetData" );
    throw( stcErr );
  }
}

///////////////////////////////////////////////////////////////////////////////
// Action
// Edit history:
// -------------
// April 12, 2005 - Created (LEM)
// November 16, 2005 - Changed equation to allow single seeds to survive and
// to take grid cell area into account (LEM)
///////////////////////////////////////////////////////////////////////////////
void clDensitySeedSurvival::Action()
{
  float fNumSeeds, fNewSeeds, fCellArea;
  int iNumXCells = mp_oSeedGrid->GetNumberXCells(), iNumYCells = mp_oSeedGrid->GetNumberYCells(), iX, iY, i;

  for ( iX = 0; iX < iNumXCells; iX++ )
  {
    if (iX == iNumXCells - 1) fCellArea = m_fXEdgeCellArea;
    else fCellArea = m_fNormalCellArea;

    for ( iY = 0; iY < iNumYCells; iY++ )
    {

      if (iY == iNumYCells - 1) {
        if (iX == iNumXCells - 1) fCellArea = m_fXYEdgeCellArea;
        else fCellArea = m_fYEdgeCellArea;
      }

      //Go through the species
      for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      {

        mp_oSeedGrid->GetValueOfCell( iX, iY, mp_iSeedGridCode[mp_iWhatSpecies[i]], & fNumSeeds );

        //Reduce the number of seeds appropriately
        if ( fNumSeeds > 1 )
        {
          fNewSeeds = clModelMath::RandomRound(fNumSeeds * exp(-mp_fDensDepSlope[mp_iWhatSpecies[i]] * pow( (fNumSeeds/fCellArea), mp_fDensDepSteepness[mp_iWhatSpecies[i]] )));
        }
        else
        {
          fNewSeeds = fNumSeeds;
        }

        if ( fNumSeeds < 0 ) fNumSeeds = 0;
        if ( fNewSeeds > fNumSeeds) fNewSeeds = fNumSeeds;

        mp_oSeedGrid->SetValueOfCell( iX, iY, mp_iSeedGridCode[mp_iWhatSpecies[i]], fNewSeeds );
      }
    }
  }
}
