//---------------------------------------------------------------------------
// FuncResponseSeedPredation.cpp
//---------------------------------------------------------------------------
#include "FuncResponseSeedPredation.h"
#include "GridBase.h"
#include "ModelMath.h"
#include "ParsingFunctions.h"
#include "SimManager.h"
#include <math.h>
#include <fstream>

/////////////////////////////////////////////////////////////////////////////
//     Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
clFuncResponseSeedPredation::clFuncResponseSeedPredation(clSimManager * p_oSimManager) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
     {
       try
       {
         strcpy(m_cNameString, "functional response seed predation");

         //Versions
         m_fVersionNumber = 1;
         m_fMinimumVersionNumber = 1;

         //Allowed file types
         m_iNumAllowedTypes = 2;
         mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
         mp_iAllowedFileTypes[0] = parfile;
         mp_iAllowedFileTypes[1] = detailed_output;

         mp_iSeedGridCode = NULL;
       }
       catch (modelErr & err)
       {
         throw(err);
       }
       catch (modelMsg & msg)
       {
         throw(msg);
       } //non-fatal error
       catch (...)
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::clFuncResponseSeedPredation");
         throw(stcErr);
       }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
clFuncResponseSeedPredation::~clFuncResponseSeedPredation()
{

  delete[] mp_iSeedGridCode;
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     Action
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clFuncResponseSeedPredation::Action()
{
  try
  {
  	using namespace std;
    fstream oOut;
    float * p_fOverallSeeds = NULL, //if outputting - number of overall seeds
         //for each timestep
         * p_fOverallPredators = NULL, //if outputting - number of overall
         //predators for each timestep
         * p_fSeedProportions = new float[m_iNumBehaviorSpecies], //proportion
         //of each species' seeds in the cell so we can preserve it
         fNumSeeds, //number of seeds in a cell
         fNumSeedsConsumed, //number of seeds consumed by predators
         fSeedFallAmount, //number new seeds added each week during seed fall
         fPerCapitaOfftake, //per capita seed offtake for each timestep
         fInstantaneousRateOfChange, //rate of change of predator population
         fPerCapitaSeedAvail, //number seeds per capita available for the week
         fOfftake, //amount of total offtake in one week
         fProportionNotGerminating = 1 - m_fProportionGerminating, fWeekToDay = 1.0 / 7.0, fNumPredators; //number of predators in a cell
    int iNumXCells = mp_oSeedGrid->GetNumberXCells(), //number X seed grid cells
         iNumYCells = mp_oSeedGrid->GetNumberYCells(), //number Y seed grid cells
         iSeedFallCounter, //counter which controls the time of seed rain
         iX, iY, //grid loop counters
         i;
    bool bOutput = false; //Whether or not to write timestep data

    if (!m_bPreservePredatorDensities)
    {
      SetPredatorInitialDensity();
    }

    //If there's a filename, open it and set the output flag to true
    if (strlen(m_cOutput) > 0)
    {
      bOutput = true;
      oOut.open(m_cOutput, ios::out | ios::app);
      oOut << "\n\nSeed predation, timestep " << mp_oSimManager->GetCurrentTimestep() << "\nWeek\tSeeds\tPredators";

      p_fOverallSeeds = new float[m_iNumWeeksToModel + 1];
      p_fOverallPredators = new float[m_iNumWeeksToModel + 1];

      for (i = 0; i < m_iNumWeeksToModel; i++)
      {
        p_fOverallSeeds[i] = 0;
        p_fOverallPredators[i] = 0;
      }
    }

    //Go through each grid cell
    for (iX = 0; iX < iNumXCells; iX++)
    {
      for (iY = 0; iY < iNumYCells; iY++)
      {

        //Get the total number of seeds and their relative proportions
        fNumSeeds = 0;
        fNumSeedsConsumed = 0;
        for (i = 0; i < m_iNumBehaviorSpecies; i++)
        {
          mp_oSeedGrid->GetValueOfCell(iX, iY, mp_iSeedGridCode[i], p_fSeedProportions + i);
          fNumSeeds += p_fSeedProportions[i];
        }
        if (fNumSeeds > 0)
        {
          for (i = 0; i < m_iNumBehaviorSpecies; i++)
          {
            p_fSeedProportions[i] /= fNumSeeds;
          }
        }

        //Get initial number of predators
        mp_oPredatorGrid->GetValueOfCell(iX, iY, m_iNumPredsCode, & fNumPredators);

        //Get the seed fall amount
        fSeedFallAmount = fNumSeeds / m_iNumWeeksSeedFall;

        //Initialize the seed fall counter and the number of seeds
        iSeedFallCounter = 1;
        fNumSeeds = 0;

        if (bOutput)
        {
          p_fOverallSeeds[0] += fSeedFallAmount;
          p_fOverallPredators[0] += fNumPredators;
        }

        //Go through the number of timesteps for this cell
        for (i = 0; i < m_iNumWeeksToModel; i++)
        {

          ////////////////////////////////////////////////////
          // Update the number of seeds
          ////////////////////////////////////////////////////

          //If this is a seed fall week, add the seed fall amount to the
          //number of seeds
          if (iSeedFallCounter <= m_iNumWeeksSeedFall)
          {
            fNumSeeds += fSeedFallAmount;
            iSeedFallCounter++;
          }

          //If this is a germination week, subtract out the number of
          //germinating seeds
          if (i >= m_iWeekGerminationStarts)
          {
            fNumSeeds *= fProportionNotGerminating;
          }

          //Get the per capita seed availability per day
          fPerCapitaSeedAvail = (fNumSeeds / fNumPredators) * fWeekToDay;

          //Get per capita seed offtake per day
          fPerCapitaOfftake = m_fMaxIntakeRate * (1 - exp(-(fPerCapitaSeedAvail * m_fForagingEfficiency)));

          //Subtract how much seed was eaten from the total seed pool - scale
          //back up to per week
          fOfftake = fPerCapitaOfftake * fNumPredators * 7;
          fNumSeeds -= fOfftake;
          fNumSeedsConsumed += fOfftake;
          if (fNumSeeds < 0) fNumSeeds = 0;

          ////////////////////////////////////////////////////
          // Update the number of predators
          ////////////////////////////////////////////////////

          //Calculate the instantaneous rate of change
          //Yes, I know fPerCapitaOfftake is per day; it's supposed to be this
          //way but I don't know why
          fInstantaneousRateOfChange =
               m_fMaxInstantaneousDeclineRate + (m_fDemographicEfficiency * fPerCapitaOfftake)
               + (m_fDensityDependentCoefficient * fNumPredators);

          //Now calculate the number of predators
          fNumPredators = fNumPredators * exp(fInstantaneousRateOfChange);

          //If number of predators is zero, stop the loop
          if (0 == fNumPredators)
          {
            break;
          }

          if (bOutput)
          {
            p_fOverallSeeds[i + 1] += fNumSeeds;
            p_fOverallPredators[i + 1] += fNumPredators;
          }
        }

        //Update the amount of seeds and predators in the grids
        for (i = 0; i < m_iNumBehaviorSpecies; i++)
        {
          mp_oSeedGrid->GetValueOfCell(iX, iY, mp_iSeedGridCode[i], & fNumSeeds);
          //Subtract from the amount of seeds the species proportion of the
          //total number consumed
          fNumSeeds -= clModelMath::RandomRound(p_fSeedProportions[i] * fNumSeedsConsumed);
          if (fNumSeeds < 0) fNumSeeds = 0;
          mp_oSeedGrid->SetValueOfCell(iX, iY, mp_iSeedGridCode[i], fNumSeeds);
        }
        mp_oPredatorGrid->SetValueOfCell(iX, iY, m_iNumPredsCode, fNumPredators);
      }
    }

    if (bOutput)
    {
      //Write out all output
      for (i = 0; i < m_iNumWeeksToModel + 1; i++)
      {
        oOut << "\n" << i << "\t" << p_fOverallSeeds[i] << "\t" << p_fOverallPredators[i];
      }
      oOut.close();
    }
    delete[] p_fSeedProportions;
    delete[] p_fOverallSeeds;
    delete[] p_fOverallPredators;

  }
  catch (modelErr & err)
  {
    throw(err);
  }
  catch (modelMsg & msg)
  {
    throw(msg);
  } //non-fatal error
  catch (...)
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::Action");
    throw(stcErr);
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     GetParameterFileData
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clFuncResponseSeedPredation::GetParameterFileData(DOMDocument * p_oDoc)
{
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();

    //Max instantaneous rate at which predator abundance declines in the
    //absense of food
    FillSingleValue(p_oElement, "pr_maxInstantaneousDeclineRate", & m_fMaxInstantaneousDeclineRate, true);

    //Population's demographic efficiency
    FillSingleValue(p_oElement, "pr_demographicEfficiency", & m_fDemographicEfficiency, true);

    //Density-dependent coefficient
    FillSingleValue(p_oElement, "pr_densityDependentCoefficient", & m_fDensityDependentCoefficient, true);

    //Maximum intake rate - number of seeds per predator per week
    FillSingleValue(p_oElement, "pr_maxIntakeRate", & m_fMaxIntakeRate, true);

    //Foraging efficiency
    FillSingleValue(p_oElement, "pr_foragingEfficiency", & m_fForagingEfficiency, true);

    //Number of weeks for which seedfall occurs
    FillSingleValue(p_oElement, "pr_numWeeksSeedFall", & m_iNumWeeksSeedFall, true);
    //Throw an error if not positive
    if (m_iNumWeeksSeedFall <= 0)
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
      strcpy(stcErr.cMoreInfo, "Number of weeks of seedfall must be greater than zero.");
      throw(stcErr);
    }

    //Number of weeks to model
    FillSingleValue(p_oElement, "pr_numWeeksToModel", & m_iNumWeeksToModel, true);
    //Throw an error if not between 0 and 53 (non-inclusive)
    if (m_iNumWeeksToModel <= 0 || m_iNumWeeksToModel > 52)
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
      strcpy(stcErr.cMoreInfo, "Number of weeks to model must be between 1 and 52 inclusive.");
      throw(stcErr);
    }

    //Week in which germination starts
    FillSingleValue(p_oElement, "pr_weekGerminationStarts", & m_iWeekGerminationStarts, true);
    //Throw an error if not between 0 and 52 (inclusive)
    if (m_iWeekGerminationStarts < 0 || m_iWeekGerminationStarts > 52)
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
      strcpy(stcErr.cMoreInfo, "Week germination starts must be between 0 and 52 inclusive.");
      throw(stcErr);
    }

    //Proportion germinating
    FillSingleValue(p_oElement, "pr_proportionGerminating", & m_fProportionGerminating, true);
    //Throw an error if not between 0 and 1 (inclusive)
    if (m_fProportionGerminating < 0 || m_fProportionGerminating > 1)
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
      strcpy(stcErr.cMoreInfo, "Proportion germinating must be between 0 and 1 inclusive.");
      throw(stcErr);
    }

    //Output filename, if desired
    FillSingleValue(p_oElement, "pr_outputFilename", m_cOutput, MAX_FILENAME_SIZE, false);

    //Whether or not to preserve predator densities
    FillSingleValue(p_oElement, "pr_preservePredatorDensities", & m_bPreservePredatorDensities, true);

    //Predator initial density in number per square meter
    FillSingleValue(p_oElement, "pr_predatorInitialDensity", & m_fPredatorInitialDensity, true);
    //Throw an error if not positive
    if (m_fPredatorInitialDensity <= 0)
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
      strcpy(stcErr.cMoreInfo, "Predator initial density must be greater than zero.");
      throw(stcErr);
    }
  }
  catch (modelErr & err)
  {
    throw(err);
  }
  catch (modelMsg & msg)
  {
    throw(msg);
  } //non-fatal error
  catch (...)
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetParameterFileData");
    throw(stcErr);
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     GetData
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clFuncResponseSeedPredation::GetData(DOMDocument * p_oDoc)
{
  try
  {
    char cLabel[MAX_GRID_LABEL_SIZE];
    int i;

    GetParameterFileData(p_oDoc);

    //Declare our array
    mp_iSeedGridCode = new short int[m_iNumBehaviorSpecies];

    //Fetch the seed grid
    mp_oSeedGrid = mp_oSimManager->GetGridObject("Dispersed Seeds");
    if (mp_oSeedGrid == NULL)
    {
      modelErr stcErr;
      strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
      strcpy(stcErr.cMoreInfo, "Disperse behaviors must be used with functional response seed predation.");
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      throw(stcErr);
    }

    //Now get the data codes
    for (i = 0; i < m_iNumBehaviorSpecies; i++)
    {
      sprintf(cLabel, "%s%d", "seeds_", mp_iWhatSpecies[i]);
      mp_iSeedGridCode[i] = mp_oSeedGrid->GetFloatDataCode(cLabel);
      if (-1 == mp_iSeedGridCode[i])
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
        strcpy(stcErr.cMoreInfo, "Unexpected grid data return code.");
        throw(stcErr);
      }
    }

    m_fCellArea = mp_oSeedGrid->GetLengthXCells() * mp_oSeedGrid->GetLengthYCells();

    //Set up the predator grid
    //Check to see if there was a map file
    mp_oPredatorGrid = mp_oSimManager->GetGridObject("Seed Predators");
    if (mp_oPredatorGrid == NULL || false == m_bPreservePredatorDensities)
    {

      //Create the grid and initialize it
      mp_oPredatorGrid = mp_oSimManager->CreateGrid("Seed Predators", //grid name
           0, //number of int data members
           1, //number of float data members
           0, //number of char data members
           0, //number of bool data members
           mp_oSeedGrid->GetLengthXCells(), //length of X cells
           mp_oSeedGrid->GetLengthYCells()); //length of Y cells

      m_iNumPredsCode = mp_oPredatorGrid->RegisterFloat("num preds");

      if (m_bPreservePredatorDensities)
      {
        SetPredatorInitialDensity();
      }
    }
    else
    {
      //Get the code - throw error if not present
      m_iNumPredsCode = mp_oPredatorGrid->GetFloatDataCode("num preds");
      if (-1 == m_iNumPredsCode)
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
        strcpy(stcErr.cMoreInfo, "Unexpected grid data return code.");
        throw(stcErr);
      }

      if (mp_oSeedGrid->GetNumberXCells() != mp_oPredatorGrid->GetNumberXCells()
           || mp_oSeedGrid->GetNumberYCells() != mp_oPredatorGrid->GetNumberYCells())
           {
             modelErr stcErr;
             stcErr.iErrorCode = BAD_DATA;
             strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
             strcpy(stcErr.cMoreInfo, "Predator grid cell resolution doesn't match seed grid cell resolution.");
             throw(stcErr);
      }
    }

  }
  catch (modelErr & err)
  {
    throw(err);
  }
  catch (modelMsg & msg)
  {
    throw(msg);
  } //non-fatal error
  catch (...)
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::GetData");
    throw(stcErr);
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     SetPredatorInitialDensity
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clFuncResponseSeedPredation::SetPredatorInitialDensity()
{
  try
  {
    float fNumPredators = m_fPredatorInitialDensity * m_fCellArea;
    int iNumXCells = mp_oPredatorGrid->GetNumberXCells(), iNumYCells = mp_oPredatorGrid->GetNumberYCells(), iX, iY;

    for (iX = 0; iX < iNumXCells; iX++)
    {
      for (iY = 0; iY < iNumYCells; iY++)
      {
        mp_oPredatorGrid->SetValueOfCell(iX, iY, m_iNumPredsCode, fNumPredators);
      }
    }
  }
  catch (modelErr & err)
  {
    throw(err);
  }
  catch (modelMsg & msg)
  {
    throw(msg);
  } //non-fatal error
  catch (...)
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy(stcErr.cFunction, "clFuncResponseSeedPredation::SetPredatorInitialDensity");
    throw(stcErr);
  }
}
//---------------------------------------------------------------------------
