//---------------------------------------------------------------------------
#include <fstream>
#include "OutputShort.h"
#include "SimManager.h"
#include "Plot.h"
#include "TreePopulation.h"
#include "ParsingFunctions.h"
#include "PlatformFuncs.h"

/////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clShortOutput::clShortOutput( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    strcpy( m_cNameString, "short output" );

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;

    //Null pointers, initialize variables to empty
    mp_treeSettings = NULL;
    mp_subplots = NULL;
    m_iNumTypesToSave = 0;
    m_iNumSubplotsToSave = 0;
    strcpy( m_cFileName, "" );

  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::clShortOutput" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 4, 2004 - Removed the try-catch block, because they shouldn't be in
// destructors( duh )( LEM )
/////////////////////////////////////////////////////////////////////////////
clShortOutput::~clShortOutput()
{
  int i; //loop counter

  //Delete arrays
  delete[] mp_treeSettings; mp_treeSettings = NULL;

  for ( i = 0; i < m_iNumSubplotsToSave; i++ )
  {
    delete[] mp_subplots[i].p_cellList; mp_subplots[i].p_cellList = NULL;
  }
  delete[] mp_subplots; mp_subplots = NULL;
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GetData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clShortOutput::GetData( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oElement; //document element
    char * cPos; //for string searching
    
    //Get the doc root element
    p_oElement = p_oDoc->getDocumentElement();

    //Get the filename
    FillSingleValue( p_oElement, "so_filename", m_cFileName, MAX_FILENAME_SIZE, true );

    //Make sure the file extension is on - if not, append it
    cPos = strstr( m_cFileName, SHORT_OUTPUT_FILE_EXT );
    if ( NULL == cPos )
    {
      //Make sure we don't exceed string length - chop it if necessary to allow
      //enough space for string concatenation
      if ( strlen( m_cFileName ) >= ( MAX_FILENAME_SIZE - strlen( SHORT_OUTPUT_FILE_EXT ) ) )
        m_cFileName[MAX_FILENAME_SIZE - strlen( SHORT_OUTPUT_FILE_EXT ) - 1] = '\0';
      strcat( m_cFileName, SHORT_OUTPUT_FILE_EXT );
    }

    //Extract tree info
    ExtractTreeInfo( p_oDoc );

    //Extract subplot info
    ExtractSubplotInfo( p_oDoc );

    //Write file header
    WriteFileHeader();

    //Write the initial conditions
    Action();

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::GetData" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// ExtractTreeInfo()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clShortOutput::ExtractTreeInfo( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    DOMNodeList * p_oTypeList; //holds the list of tree type info structures
    DOMNode * p_oDocNode;
    DOMElement * p_oTreeInfo; //one tree type's info
    short int iNumTypes, //how many types we have output data for
         i; //loop counter
    char * cData; //extracting text data


    //****************************
    //  Look for tree info
    //****************************
    //Search for all tree type info sets
    p_oTypeList = p_oDoc->getElementsByTagName( XMLString::transcode( "so_treeTypeInfo" ) );
    iNumTypes = p_oTypeList->getLength();

    //If we found nothing, exit with no error
    if ( iNumTypes == 0 ) return;

    //Allocate the grid memory and initialize
    m_iNumTypesToSave = iNumTypes;
    mp_treeSettings = new stcTreeOutputInfo[m_iNumTypesToSave];
    for ( i = 0; i < m_iNumTypesToSave; i++ )
    {
      mp_treeSettings[i].bSaveRBA = false;
      mp_treeSettings[i].bSaveABA = false;
      mp_treeSettings[i].bSaveRDN = false;
      mp_treeSettings[i].bSaveADN = false;
    }

    //Go through each type and break out its info
    for ( i = 0; i < m_iNumTypesToSave; i++ )
    {
      p_oDocNode = p_oTypeList->item( i );
      p_oTreeInfo = ( DOMElement * ) p_oDocNode;

      //get the type name - if something goes wrong throw an error
      cData = XMLString::transcode( p_oTreeInfo->getAttributeNode( XMLString::transcode( "type" ) )->getNodeValue() );
      if ( strcmp( "Seedling", cData ) == 0 )
        mp_treeSettings[i].iType = clTreePopulation::seedling;
      else if ( strcmp( "Sapling", cData ) == 0 )
        mp_treeSettings[i].iType = clTreePopulation::sapling;
      else if ( strcmp( "Adult", cData ) == 0 )
        mp_treeSettings[i].iType = clTreePopulation::adult;
      else if ( strcmp( "Snag", cData ) == 0 )
        mp_treeSettings[i].iType = clTreePopulation::snag;
      else
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clShortOutput::ExtractTreeInfo" );
        stcErr.iErrorCode = BAD_DATA;
        sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized or unsupported type - ", cData );
        throw( stcErr );
      }

      //Check the data save tags
      mp_treeSettings[i].bSaveRBA = TestForSave( p_oTreeInfo, "so_saveRBA" );
      mp_treeSettings[i].bSaveABA = TestForSave( p_oTreeInfo, "so_saveABA" );
      mp_treeSettings[i].bSaveRDN = TestForSave( p_oTreeInfo, "so_saveRDN" );
      mp_treeSettings[i].bSaveADN = TestForSave( p_oTreeInfo, "so_saveADN" );

      //If this is a seedling, turn off options for basal area
      if ( clTreePopulation::seedling == mp_treeSettings[i].iType )
      {
        mp_treeSettings[i].bSaveRBA = false;
        mp_treeSettings[i].bSaveABA = false;
      }
    } //end of for (i = 0; i < m_iNumTypeToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::ExtractTreeInfo" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////
// ExtractSubplotInfo()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
////////////////////////////////////////////////////////////////////////////
void clShortOutput::ExtractSubplotInfo( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
    stcCoords * p_gridListForVal; //this is a list of grid cells that we will
    //validate before putting in the object list
    DOMNodeList * p_oGridCellList, //for getting the list of grid cells to save for
         * p_oSubplotList; //list of subplots
    DOMNode * p_oDocNode; //for retrieving data
    DOMElement * p_oSubplot, //one subplot element
         * p_oElement; //generic element object
    int iX, iY; //grid cell X and Y coordinate numbers
    short int iNumSubplots, //number of subplots found
         iNumCells, //number of grid cells found for one subplot
         iNumPlotXCells = p_oPop->GetNumXCells(), 
         iNumPlotYCells = p_oPop->GetNumYCells(), i, j, k, m; //loop counters
    bool bDuplicate; //for searching for duplicate values

    //Get all the subplot info structures out of the file, if any
    p_oSubplotList = p_oDoc->getElementsByTagName( XMLString::transcode( "so_subplot" ) );
    iNumSubplots = p_oSubplotList->getLength();
    if ( 0 == iNumSubplots ) return; //if nothing found exit without error

    //Allocate memory
    m_iNumSubplotsToSave = iNumSubplots;
    mp_subplots = new stcSubplotInfo[m_iNumSubplotsToSave];
    for ( i = 0; i < m_iNumSubplotsToSave; i++ )
    {
      mp_subplots[i].p_cellList = NULL;
      mp_subplots[i].iNumCells = 0;
    } //end of for (i = 0; i < m_iNumSubplotsToSave; i++)

    //Go through the subplots in the file one at a time and extract the info
    for ( i = 0; i < m_iNumSubplotsToSave; i++ )
    {
      p_oDocNode = p_oSubplotList->item( i );
      p_oSubplot = ( DOMElement * ) p_oDocNode;

      //Get subplot name
      FillSingleValue( p_oSubplot, "so_subplotName", mp_subplots[i].cSubplotName, MAX_SUBPLOT_NAME_SIZE, true );

      //Get grid coords
      p_oGridCellList = p_oSubplot->getElementsByTagName( XMLString::transcode( "pointSet" ) );
      if ( 0 == p_oGridCellList->getLength() )
      {
        modelErr stcErr;
        stcErr.iErrorCode = DATA_MISSING;
        strcpy( stcErr.cFunction, "clShortOutput::ExtractSubplotInfo" );
        strcpy( stcErr.cMoreInfo, "pointSet" );
        throw( stcErr );
      }
      p_oDocNode = p_oGridCellList->item( 0 );
      p_oElement = ( DOMElement * ) p_oDocNode;
      p_oGridCellList = p_oElement->getElementsByTagName( XMLString::transcode( "po_point" ) );
      iNumCells = p_oGridCellList->getLength();
      if ( 0 == iNumCells )
      {
        modelErr stcErr;
        stcErr.iErrorCode = DATA_MISSING;
        strcpy( stcErr.cFunction, "clShortOutput::ExtractSubplotInfo" );
        strcpy( stcErr.cMoreInfo, "po_point" );
        throw( stcErr );
      }

      //Put the grid coords in a temp list so we can validate them
      p_gridListForVal = new stcCoords[iNumCells];
      k = 0;
      for ( j = 0; j < iNumCells; j++ )
      {
        p_oDocNode = p_oGridCellList->item( j );
        p_oElement = ( DOMElement * ) p_oDocNode;
        iX = atoi( XMLString::transcode( p_oElement->getAttributeNode( XMLString::transcode( "x" ) )->getNodeValue() ) );
        iY = atoi( XMLString::transcode( p_oElement->getAttributeNode( XMLString::transcode( "y" ) )->getNodeValue() ) );
        //Make sure that the values are within the plot
        if ( iX >= iNumPlotXCells || iX < 0 || iY >= iNumPlotYCells || iY < 0 )
        {
          modelErr stcErr;
          strcpy( stcErr.cFunction, "clShortOutput::ExtractSubplotInfo" );
          sprintf( stcErr.cMoreInfo, "%s%d%s%d", "Point value outside plot - ", iX, ", ", iY );
          stcErr.iErrorCode = BAD_DATA;
        }
        //Make sure they're not duplicates - this is not cause to abort if
        //they are
        bDuplicate = false;
        for ( m = 0; m < j; m++ )
          if ( iX == p_gridListForVal[m].iX && iY == p_gridListForVal[m].iY )
            bDuplicate = true;
        if ( !bDuplicate )
        { //our values are valid and not duplicates
          p_gridListForVal[k].iX = iX;
          p_gridListForVal[k].iY = iY;
          k++;
        }
        else
          iNumCells--; //duplicate - decrease the count by one
      } //end of for (j = 0; j < iNumCells; j++)

      //transfer the no-duplicates list over to our object variables
      if ( iNumCells > 0 )
      {
        mp_subplots[i].p_cellList = new stcCoords[iNumCells];
        mp_subplots[i].iNumCells = iNumCells;
        for ( j = 0; j < iNumCells; j++ )
        {
          mp_subplots[i].p_cellList[j].iX = p_gridListForVal[j].iX;
          mp_subplots[i].p_cellList[j].iY = p_gridListForVal[j].iY;
        }
      }
      delete[] p_gridListForVal;

    } //end of for (i = 0; i < m_iNumSubplotsToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::ExtractSubplotInfo" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// TestForSave()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
bool clShortOutput::TestForSave( DOMElement * p_oParentElement, char * cNodeName )
{
  try
  {
    DOMNodeList * p_oNodeList; //for searching by tag name
    DOMNode * p_oDocNode; //for retrieving one node of the search
    DOMElement * p_oElement; //for casting a DOM_Node to a DOMElement
    char * cData; //for extracting the attribute text

    p_oNodeList = p_oParentElement->getElementsByTagName( XMLString::transcode( cNodeName ) );
    if ( 0 == p_oNodeList->getLength() ) return false; //nothing found

    p_oDocNode = p_oNodeList->item( 0 );
    p_oElement = ( DOMElement * ) p_oDocNode;
    cData = XMLString::transcode( p_oElement->getAttributeNode( XMLString::transcode( "save" ) )->getNodeValue() );
    if ( strcmp( "true", cData ) == 0 ) return true;

    return false;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::TestForSave" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// WriteFileHeader()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// June 24, 2004 - Added snags and changed column headers to be more meaningful
// October 21, 2004 - Made it so a new run's data doesn't append to an old file
//                    but instead overwrites it (LEM)
// April 24, 2007 - Added a check to throw an error if we don't have write
// permission (LEM)
/////////////////////////////////////////////////////////////////////////////
void clShortOutput::WriteFileHeader()
{
  try
  {
  	using namespace std;
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
    clPlot * p_oPlot = mp_oSimManager->GetPlotObject();
    fstream out( m_cFileName, ios::out | ios::trunc );
    char cABAString[20], cABATotalString[20], cRBAString[20], cADNString[20], cADNTotalString[20], cRDNString[20];
    short int iNumSpecies = p_oPop->GetNumberOfSpecies(), i, j; //loop counter

    out << "Location: " << p_oPlot->GetPlotTitle() << "\n" 
        << "From parameter file: " << mp_oSimManager->GetParFilename() << "\n" 
        << "Basal area is normalized to m^2/ha; density to #/ha.\n" 
        << "Subplots:\n0 = Whole plot\n";

    //Write subplot list
    for ( i = 0; i < m_iNumSubplotsToSave; i++ )
      out << i + 1 << " = " << mp_subplots[i].cSubplotName << "\n";

    //Now write column headers
    out << "Step\tSubplot";

    for ( i = 0; i < m_iNumTypesToSave; i++ )
    {

      //Set up column header strings
      if ( clTreePopulation::seedling == mp_treeSettings[i].iType )
      {

        strcpy( cADNString, "Sdl Abs Den: " );
        strcpy( cADNTotalString, "Sdl Den Total:" );
        strcpy( cRDNString, "Sdl Rel Den: " );

      }
      else if ( clTreePopulation::sapling == mp_treeSettings[i].iType )
      {

        strcpy( cABAString, "Sapl Abs BA: " );
        strcpy( cABATotalString, "Sapl BA Total:" );
        strcpy( cRBAString, "Sapl Rel BA: " );
        strcpy( cADNString, "Sapl Abs Den: " );
        strcpy( cADNTotalString, "Sapl Den Total:" );
        strcpy( cRDNString, "Sapl Rel Den: " );

      }
      else if ( clTreePopulation::adult == mp_treeSettings[i].iType )
      {

        strcpy( cABAString, "Adult Abs BA: " );
        strcpy( cABATotalString, "Adult BA Total:" );
        strcpy( cRBAString, "Adult Rel BA: " );
        strcpy( cADNString, "Adult Abs Den: " );
        strcpy( cADNTotalString, "Adult Den Total:" );
        strcpy( cRDNString, "Adult Rel Den: " );

      }
      else if ( clTreePopulation::snag == mp_treeSettings[i].iType )
      {

        strcpy( cABAString, "Snag Abs BA: " );
        strcpy( cABATotalString, "Snag BA Total:" );
        strcpy( cRBAString, "Snag Rel BA: " );
        strcpy( cADNString, "Snag Abs Den: " );
        strcpy( cADNTotalString, "Snag Den Total:" );
        strcpy( cRDNString, "Snag Rel Den: " );

      }

      if ( mp_treeSettings[i].bSaveABA )
      {
        for ( j = 0; j < iNumSpecies; j++ )
          out << "\t" << cABAString << p_oPop->TranslateSpeciesCodeToName( j );

        out << "\t" << cABATotalString;
      }

      if ( mp_treeSettings[i].bSaveRBA )
        for ( j = 0; j < iNumSpecies; j++ )
          out << "\t" << cRBAString << p_oPop->TranslateSpeciesCodeToName( j );

      if ( mp_treeSettings[i].bSaveADN )
      {
        for ( j = 0; j < iNumSpecies; j++ )
          out << "\t" << cADNString << p_oPop->TranslateSpeciesCodeToName( j );

        out << "\t" << cADNTotalString;
      }

      if ( mp_treeSettings[i].bSaveRDN )
        for ( j = 0; j < iNumSpecies; j++ )
          out << "\t" << cRDNString << p_oPop->TranslateSpeciesCodeToName( j );
    }

    out << "\n";
    out.close();
    
    //As a check, see if the file exists. If not, throw an error
    //because we didn't have write permission.
    if (!DoesFileExist(m_cFileName)) {
      modelErr stcErr;
      stcErr.iErrorCode = DATA_MISSING;
      strcpy( stcErr.cFunction, "clShortOutput::WriteFileHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "SORTIE could not write the file \"",
              m_cFileName, "\". Check for invalid directories." );
      throw( stcErr );
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::WriteFileHeader" );
    throw( stcErr );
  }
}


/////////////////////////////////////////////////////////////////////////////
// Action()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clShortOutput::Action()
{
  try
  {
  	using namespace std;
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreePopulation * p_oTrees = ( clTreePopulation * ) p_oTemp;
    clPlot * p_oPlot = mp_oSimManager->GetPlotObject();
    clTreeSearch * p_oTypeTrees; //pointer to search results after type search
    clTree * p_oTree; //for getting trees from the search object
    fstream out( m_cFileName, ios::app | ios::out );
    float fDbh, //dbh of tree if not a seed or seedling
         fTotal, //for relativizing values
         fBA, //tree's basal area
         fCoord, //tree's X or Y coordinate
         fPlotArea = p_oPlot->GetPlotArea(), //area of the plot, in hectares
         * * p_fRBA = NULL, //array for relative basal area - #species by #types
         * * p_fABA = NULL, //array for absolute basal area - #species by #types
         * * p_fRDN = NULL, //array for relative density - #species by #types
         * * p_fADN = NULL, //array for absolute density - #species by #types
         * * * p_fSubRBA = NULL, //subplots
         * * * p_fSubABA = NULL, * * * p_fSubRDN = NULL, * * * p_fSubADN = NULL, * p_fSubplotArea = NULL; //array for subplot area in hectares
    bool * p_bUseThisType, * p_bUseBA;
    int iNumSpecies = p_oTrees->GetNumberOfSpecies(), 
        iNumTypes = p_oTrees->GetNumberOfTypes(), 
        iGridCellSize = p_oTrees->GetGridCellSize(), iTreeX, iTreeY, //tree's
         // X and Y grid cell numbers
         iSpecies, //tree's species
         iNumCells = p_oTrees->GetNumXCells() * p_oTrees->GetNumYCells(), i, j, k, iTypeIndex, iSubplotIndex; //loop counters

    //Set up flags for whether each type is to be saved
    p_bUseThisType = new bool[iNumTypes];
    p_bUseBA = new bool[iNumTypes];
    for ( i = 0; i < iNumTypes; i++ )
    {
      p_bUseThisType[i] = false;
      p_bUseBA[i] = false;
    }

    for ( i = 0; i < m_iNumTypesToSave; i++ )
    {
      if ( mp_treeSettings[i].bSaveRBA || mp_treeSettings[i].bSaveABA )
        p_bUseBA[mp_treeSettings[i].iType] = true;
      if ( p_bUseBA[mp_treeSettings[i].iType] || 
           mp_treeSettings[i].bSaveADN || 
           mp_treeSettings[i].bSaveRDN )
        p_bUseThisType[mp_treeSettings[i].iType] = true;
    }

    //Declare our data arrays
    p_fRBA = new float * [iNumTypes];
    p_fABA = new float * [iNumTypes];
    p_fRDN = new float * [iNumTypes];
    p_fADN = new float * [iNumTypes];
    for ( i = 0; i < iNumTypes; i++ )
    {
      p_fRBA[i] = new float[iNumSpecies];
      p_fABA[i] = new float[iNumSpecies];
      p_fRDN[i] = new float[iNumSpecies];
      p_fADN[i] = new float[iNumSpecies];
      for ( j = 0; j < iNumSpecies; j++ )
      {
        p_fRBA[i] [j] = 0;
        p_fABA[i] [j] = 0;
        p_fRDN[i] [j] = 0;
        p_fADN[i] [j] = 0;
      }
    }

    if ( m_iNumSubplotsToSave > 0 )
    {
      p_fSubRBA = new float * * [m_iNumSubplotsToSave];
      p_fSubABA = new float * * [m_iNumSubplotsToSave];
      p_fSubRDN = new float * * [m_iNumSubplotsToSave];
      p_fSubADN = new float * * [m_iNumSubplotsToSave];
      for ( i = 0; i < m_iNumSubplotsToSave; i++ )
      {
        p_fSubRBA[i] = new float * [iNumTypes];
        p_fSubABA[i] = new float * [iNumTypes];
        p_fSubRDN[i] = new float * [iNumTypes];
        p_fSubADN[i] = new float * [iNumTypes];
        for ( j = 0; j < iNumTypes; j++ )
        {
          p_fSubRBA[i] [j] = new float[iNumSpecies];
          p_fSubABA[i] [j] = new float[iNumSpecies];
          p_fSubRDN[i] [j] = new float[iNumSpecies];
          p_fSubADN[i] [j] = new float[iNumSpecies];
          for ( k = 0; k < iNumSpecies; k++ )
          {
            p_fSubRBA[i] [j] [k] = 0;
            p_fSubABA[i] [j] [k] = 0;
            p_fSubRDN[i] [j] [k] = 0;
            p_fSubADN[i] [j] [k] = 0;
          }
        }
      }

      p_fSubplotArea = new float[m_iNumSubplotsToSave];
      for ( i = 0; i < m_iNumSubplotsToSave; i++ )
      {

        //For each subplot calculate the area in hectares
        /** @todo fix this - this is not an accurate calculation if 8 does not divide evenly into the plot lengths */
        p_fSubplotArea[i] = ( float )mp_subplots[i].iNumCells / ( float )iNumCells * fPlotArea;

        //Invert it
        if ( p_fSubplotArea[i] > 0 )
          p_fSubplotArea[i] = 1 / p_fSubplotArea[i];
      }
    }


    //******************************
    // Tree data saving
    //******************************

    //Get all trees of this type
    p_oTypeTrees = p_oTrees->Find( "all" );

    //Go through the trees one at a time and add each's data to the running
    //totals
    p_oTree = p_oTypeTrees->NextTree();
    while ( p_oTree )
    {
      iTypeIndex = p_oTree->GetType();

      if ( p_bUseThisType[iTypeIndex] )
      {
        iSpecies = p_oTree->GetSpecies();

        //Get the tree's X and Y grid cell numbers
        p_oTree->GetValue( p_oTrees->GetXCode( iSpecies, iTypeIndex ), & fCoord );
        iTreeX = (int)floor( fCoord / iGridCellSize );
        p_oTree->GetValue( p_oTrees->GetYCode( iSpecies, iTypeIndex ), & fCoord );
        iTreeY = (int)floor( fCoord / iGridCellSize );

        if ( p_bUseBA[iTypeIndex] )
        {

          //Basal area
          //Get the dbh
          p_oTree->GetValue( p_oTrees->GetDbhCode( iSpecies, iTypeIndex ), & fDbh );

          //Calculate basal area and add to array total
          fBA = clModelMath::CalculateBasalArea( fDbh );
          p_fABA[iTypeIndex] [iSpecies] += fBA;

        } //end of check for basal area

        //Density
        p_fADN[iTypeIndex] [iSpecies] ++;

        //*****************************
        // Subplots
        //*****************************
        //Check to see if this tree is in any subplots
        for ( iSubplotIndex = 0; iSubplotIndex < m_iNumSubplotsToSave; iSubplotIndex++ )
        {
          for ( i = 0; i < mp_subplots[iSubplotIndex].iNumCells; i++ )
            if ( iTreeX == mp_subplots[iSubplotIndex].p_cellList[i].iX
                 && iTreeY == mp_subplots[iSubplotIndex].p_cellList[i].iY )
                 {

                   if ( p_bUseBA[iTypeIndex] )
                     p_fSubABA[iSubplotIndex] [iTypeIndex] [iSpecies] += fBA;

                   p_fSubADN[iSubplotIndex] [iTypeIndex] [iSpecies] ++;
            }
        }
      }

      p_oTree = p_oTypeTrees->NextTree();
    } //end of while (p_oTree)

    for ( iTypeIndex = 0; iTypeIndex < m_iNumTypesToSave; iTypeIndex++ )
    {

      //*******************************
      // Calculate relative values, where applicable
      //*******************************
      if ( mp_treeSettings[iTypeIndex].bSaveRBA )
      {

        fTotal = 0;

        //Add up all the basal area for all the species
        for ( i = 0; i < iNumSpecies; i++ )
          fTotal += p_fABA[mp_treeSettings[iTypeIndex].iType] [i];

        //Divide species values by total and put in relative BA array
        for ( i = 0; i < iNumSpecies; i++ )
        {
          if ( fTotal > 0 )
          {
            p_fRBA[mp_treeSettings[iTypeIndex].iType] [i] = p_fABA[mp_treeSettings[iTypeIndex].iType] [i] / fTotal;
          }
          else
          {
            p_fRBA[mp_treeSettings[iTypeIndex].iType] [i] = 0;
          }
        }

        //Now do subplots
        for ( iSubplotIndex = 0; iSubplotIndex < m_iNumSubplotsToSave; iSubplotIndex++ )
        {

          fTotal = 0;
          for ( i = 0; i < iNumSpecies; i++ )
            fTotal += p_fSubABA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];

          for ( i = 0; i < iNumSpecies; i++ )
          {
            if ( fTotal > 0 )
            {
              p_fSubRBA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i] =
                   p_fSubABA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i] / fTotal;
            }
            else
            {
              p_fSubRBA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i] = 0;
            }
          }
        }
      } //end of if (mp_treeSettings[iType].bSaveRBA)

      if ( mp_treeSettings[iTypeIndex].bSaveRDN )
      {

        fTotal = 0;

        //Add up all the density for all the species
        for ( i = 0; i < iNumSpecies; i++ )
          fTotal += p_fADN[mp_treeSettings[iTypeIndex].iType] [i];
        //Divide species values by total and put in relative den array
        for ( i = 0; i < iNumSpecies; i++ )
        {

          if ( fTotal > 0 )
          {
            p_fRDN[mp_treeSettings[iTypeIndex].iType] [i] = p_fADN[mp_treeSettings[iTypeIndex].iType] [i] / fTotal;
          }
          else
          {
            p_fRDN[mp_treeSettings[iTypeIndex].iType] [i] = 0;
          }
        }

        //Now do subplots
        for ( iSubplotIndex = 0; iSubplotIndex < m_iNumSubplotsToSave; iSubplotIndex++ )
        {

          fTotal = 0;
          for ( i = 0; i < iNumSpecies; i++ )
            fTotal += p_fSubADN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];

          for ( i = 0; i < iNumSpecies; i++ )
          {
            if ( fTotal > 0 )
            {
              p_fSubRDN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i] =
                   p_fSubADN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i] / fTotal;
            }
            else
            {
              p_fSubRDN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i] = 0;
            }
          }
        }
      } //end of if (mp_treeSettings[iType].bSaveRDN)

      // Convert all absolute values to numbers per hectare by dividing by plot area
      if ( mp_treeSettings[iTypeIndex].bSaveABA )
      {
        for ( i = 0; i < iNumSpecies; i++ )
          p_fABA[mp_treeSettings[iTypeIndex].iType] [i] /= fPlotArea;

        for ( i = 0; i < m_iNumSubplotsToSave; i++ )
          for ( j = 0; j < iNumSpecies; j++ )
            p_fSubABA[i] [mp_treeSettings[iTypeIndex].iType] [j] *= p_fSubplotArea[i];
      }

      if ( mp_treeSettings[iTypeIndex].bSaveADN )
      {
        for ( i = 0; i < iNumSpecies; i++ )
          p_fADN[mp_treeSettings[iTypeIndex].iType] [i] /= fPlotArea;

        for ( i = 0; i < m_iNumSubplotsToSave; i++ )
          for ( j = 0; j < iNumSpecies; j++ )
            p_fSubADN[i] [mp_treeSettings[iTypeIndex].iType] [j] *= p_fSubplotArea[i];
      }
    } //end of for (iTypeIndex = 0; iTypeIndex < m_iNumTypesToSave; iTypeIndex++)

    //Write data to file
    out << mp_oSimManager->GetCurrentTimestep() << "\t0"; //zeroth subplot = whole
    //plot
    for ( iTypeIndex = 0; iTypeIndex < m_iNumTypesToSave; iTypeIndex++ )
    {

      if ( mp_treeSettings[iTypeIndex].bSaveABA )
      {
        fTotal = 0;
        for ( i = 0; i < iNumSpecies; i++ )
        {
          fTotal += p_fABA[mp_treeSettings[iTypeIndex].iType] [i];
          out << "\t" << p_fABA[mp_treeSettings[iTypeIndex].iType] [i];
        }
        //Total line
        out << "\t" << fTotal;
      }

      if ( mp_treeSettings[iTypeIndex].bSaveRBA )
        for ( i = 0; i < iNumSpecies; i++ )
          out << "\t" << p_fRBA[mp_treeSettings[iTypeIndex].iType] [i];

      if ( mp_treeSettings[iTypeIndex].bSaveADN )
      {
        fTotal = 0;
        for ( i = 0; i < iNumSpecies; i++ )
        {
          fTotal += p_fADN[mp_treeSettings[iTypeIndex].iType] [i];
          out << "\t" << p_fADN[mp_treeSettings[iTypeIndex].iType] [i];
        }
        //Total line
        out << "\t" << fTotal;
      }

      if ( mp_treeSettings[iTypeIndex].bSaveRDN )
        for ( i = 0; i < iNumSpecies; i++ )
          out << "\t" << p_fRDN[mp_treeSettings[iTypeIndex].iType] [i];

    } //end of for (iTypeIndex = 0; iTypeIndex < m_iNumTypesToSave; iTypeIndex++)
    out << "\n";

    //Now subplots
    for ( iSubplotIndex = 0; iSubplotIndex < m_iNumSubplotsToSave; iSubplotIndex++ )
    {
      out << mp_oSimManager->GetCurrentTimestep() << "\t" << iSubplotIndex + 1;

      for ( iTypeIndex = 0; iTypeIndex < m_iNumTypesToSave; iTypeIndex++ )
      {

        if ( mp_treeSettings[iTypeIndex].bSaveABA )
        {
          fTotal = 0;
          for ( i = 0; i < iNumSpecies; i++ )
          {
            fTotal += p_fSubABA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];
            out << "\t" << p_fSubABA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];
          }
          //Total line
          out << "\t" << fTotal;
        }

        if ( mp_treeSettings[iTypeIndex].bSaveRBA )
          for ( i = 0; i < iNumSpecies; i++ )
            out << "\t" << p_fSubRBA[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];

        if ( mp_treeSettings[iTypeIndex].bSaveADN )
        {
          fTotal = 0;
          for ( i = 0; i < iNumSpecies; i++ )
          {
            fTotal += p_fSubADN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];
            out << "\t" << p_fSubADN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];
          }
          //Total line
          out << "\t" << fTotal;
        }

        if ( mp_treeSettings[iTypeIndex].bSaveRDN )
          for ( i = 0; i < iNumSpecies; i++ )
            out << "\t" << p_fSubRDN[iSubplotIndex] [mp_treeSettings[iTypeIndex].iType] [i];

      } //end of for (iTypeIndex = 0; iTypeIndex < m_iNumTypesToSave; iTypeIndex++)

      out << "\n";
    }

    out.close();

    //Delete arrays
    delete[] p_bUseThisType;
    delete[] p_bUseBA;

    for ( i = 0; i < iNumTypes; i++ )
    {
      delete[] p_fABA[i];
      delete[] p_fRBA[i];
      delete[] p_fADN[i];
      delete[] p_fRDN[i];
    }
    delete[] p_fABA;
    delete[] p_fRBA;
    delete[] p_fADN;
    delete[] p_fRDN;

    for ( i = 0; i < m_iNumSubplotsToSave; i++ )
    {
      for ( j = 0; j < iNumTypes; j++ )
      {
        delete[] p_fSubRBA[i] [j];
        delete[] p_fSubABA[i] [j];
        delete[] p_fSubRDN[i] [j];
        delete[] p_fSubADN[i] [j];
      }
      delete[] p_fSubRBA[i];
      delete[] p_fSubABA[i];
      delete[] p_fSubRDN[i];
      delete[] p_fSubADN[i];
    }
    delete[] p_fSubRBA;
    delete[] p_fSubABA;
    delete[] p_fSubRDN;
    delete[] p_fSubADN;
    delete[] p_fSubplotArea;

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clShortOutput::Action" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------



